"""Base classes for backend implementations."""

from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any

import libcst as cst

if TYPE_CHECKING:
    from senzo.settings import EnumStyle, HooksConfig


@dataclass
class FieldDefinition:
    """Definition of a field in a generated type."""

    name: str
    api_name: str
    type_annotation: str
    required: bool
    default: Any | None = None
    description: str | None = None


@dataclass
class TypeMapping:
    """Mapping from OpenAPI types to Python types."""

    string: str = "str"
    integer: str = "int"
    number: str = "float"
    boolean: str = "bool"
    date: str = "datetime.date"
    datetime: str = "datetime.datetime"
    uuid: str = "uuid.UUID"
    email: str = "str"
    uri: str = "str"
    binary: str = "bytes"
    decimal: str = "decimal.Decimal"
    formats: dict[str, str] = field(default_factory=lambda: {})

    def get_format_type(self, openapi_type: str, format_: str | None) -> str | None:
        """Get the Python type for a specific format."""
        if format_ is None:
            return None
        key = f"{openapi_type}:{format_}"
        if key in self.formats:
            return self.formats[key]
        format_map = {
            "string:date": self.date,
            "string:date-time": self.datetime,
            "string:uuid": self.uuid,
            "string:email": self.email,
            "string:uri": self.uri,
            "string:binary": self.binary,
            "number:decimal": self.decimal,
        }
        return format_map.get(key)


class DataclassBackend(ABC):
    """Base class for dataclass generation backends."""

    def __init__(self, enum_style: EnumStyle | None = None) -> None:
        """Initialize backend with enum style configuration."""
        from senzo.settings import EnumStyle as ES

        self._enum_style = enum_style if enum_style is not None else ES.ENUM

    @property
    def enum_style(self) -> EnumStyle:
        """Get the enum style for this backend."""
        return self._enum_style

    @abstractmethod
    def generate_class(
        self,
        name: str,
        fields: list[FieldDefinition],
        docstring: str | None = None,
    ) -> cst.ClassDef:
        """Generate a dataclass definition."""
        ...

    @abstractmethod
    def generate_enum(
        self,
        name: str,
        values: list[tuple[str, Any]],
        style: EnumStyle | None = None,
    ) -> cst.ClassDef:
        """Generate an enum definition.

        Args:
            name: The enum class name
            values: List of (name, value) tuples for enum members
            style: Override enum style, or use backend default
        """
        ...

    def generate_literal_type(self, values: list[Any]) -> str:
        """Generate a Literal type annotation for enum values.

        Args:
            values: The enum values

        Returns:
            A Literal type string like 'Literal["a", "b", "c"]'
        """
        literal_values = ", ".join(repr(v) for v in values)
        return f"Literal[{literal_values}]"

    @abstractmethod
    def get_imports(self) -> list[cst.SimpleStatementLine]:
        """Return required imports for this backend."""
        ...

    @abstractmethod
    def get_type_mapping(self) -> TypeMapping:
        """Return type mappings for OpenAPI types."""
        ...

    @abstractmethod
    def supports_type(self, openapi_type: str, format_: str | None) -> bool:
        """Check if backend supports a specific type."""
        ...

    @abstractmethod
    def generate_decode_json_expression(
        self,
        type_name: str,
        bytes_expr: cst.BaseExpression,
    ) -> cst.BaseExpression:
        """Generate an expression to decode JSON bytes to the given type.

        Args:
            type_name: The name of the type to decode to (e.g., "PostResponse")
            bytes_expr: A CST expression that evaluates to bytes (e.g., response.content)

        Returns:
            A CST expression that decodes the bytes to the typed object
        """
        ...

    @abstractmethod
    def generate_encode_json_expression(
        self,
        type_annotation: str,
        body_expr: cst.BaseExpression,
    ) -> cst.BaseExpression:
        """Generate an expression to encode an object to JSON bytes.

        Args:
            type_annotation: The type annotation string for the body (e.g., "dict[str, Any]")
            body_expr: A CST expression for the object to encode

        Returns:
            A CST expression that encodes the object to JSON bytes
        """
        ...

    @abstractmethod
    def get_client_imports(self) -> list[cst.SimpleStatementLine]:
        """Return imports needed in client.py for encoding/decoding."""
        ...


class HttpBackend(ABC):
    """Base class for HTTP client generation."""

    def __init__(self, async_mode: bool = False) -> None:
        self.async_mode = async_mode
        self._dataclass_backend: DataclassBackend | None = None
        self._hooks_config: HooksConfig | None = None

    def set_dataclass_backend(self, backend: DataclassBackend) -> None:
        """Set the dataclass backend for response decoding."""
        self._dataclass_backend = backend

    def set_hooks_config(self, config: HooksConfig | None) -> None:
        """Set the hooks configuration for code generation."""
        self._hooks_config = config

    @abstractmethod
    def generate_client_class(
        self,
        name: str,
        base_url_param: bool = True,
    ) -> cst.ClassDef:
        """Generate the client class with constructor."""
        ...

    @abstractmethod
    def generate_sub_client_class(
        self,
        name: str,
    ) -> cst.ClassDef:
        """Generate a sub-client class that shares an existing session.

        Sub-clients accept an existing HTTP client/session instead of creating
        their own. They do not manage lifecycle (no close(), no context manager
        methods) - that is handled by the parent client.
        """
        ...

    @abstractmethod
    def generate_method(
        self,
        operation: Any,
    ) -> cst.FunctionDef:
        """Generate an API method."""
        ...

    @abstractmethod
    def get_imports(self) -> list[cst.SimpleStatementLine]:
        """Return required imports."""
        ...

    @abstractmethod
    def supports_async(self) -> bool:
        """Whether this backend supports async."""
        ...

    @abstractmethod
    def supports_sync(self) -> bool:
        """Whether this backend supports sync."""
        ...

    def supports_websocket(self) -> bool:
        """Return True if this backend supports WebSocket connections.

        Default is False. Backends that support WebSocket should override this.
        """
        return False

    def get_base_import_names(self) -> list[str]:
        """Return additional names to import from _base.py.

        Override in subclasses that need extra imports like Handler, Sequence, etc.
        """
        return []

    def get_stdlib_imports(self) -> list[cst.SimpleStatementLine]:
        """Return standard library imports needed (e.g., collections.abc).

        Override in subclasses that need stdlib imports.
        """
        return []

    def generate_websocket_method(
        self,
        operation: Any,
    ) -> cst.FunctionDef:
        """Generate a WebSocket connection method.

        Args:
            operation: The WebSocketOperation to generate a method for.

        Returns:
            A CST FunctionDef for the WebSocket connection method.

        Raises:
            NotImplementedError: If the backend doesn't support WebSocket.
        """
        raise NotImplementedError(
            f"{self.__class__.__name__} does not support WebSocket"
        )

    def get_websocket_runtime_code(self) -> str | None:
        """Return runtime code for WebSocket support.

        This includes the WebSocketConnection class and related types.
        Returns None if the backend doesn't support WebSocket.
        """
        return None

    def get_sub_client_init_args(self) -> list[tuple[str, str]]:
        """Return the (param_name, attr_expr) pairs for sub-client __init__ calls.

        Used by the generator to build sub-client initialization in the main client.
        For example, httpx returns [("client", "self._client")] meaning sub-clients
        are initialized with `SubClient(client=self._client)`.

        Returns:
            List of (parameter_name, expression_string) tuples.
        """
        raise NotImplementedError(
            f"{self.__class__.__name__} must implement get_sub_client_init_args"
        )



"""LibCST utilities for code generation."""

from __future__ import annotations

from typing import Sequence

import libcst as cst


def make_name(name: str) -> cst.Name:
    """Create a Name node."""
    return cst.Name(name)


def make_annotation(type_str: str) -> cst.Annotation:
    """Create an Annotation node from a type string."""
    return cst.Annotation(annotation=cst.parse_expression(type_str))


def make_simple_string(value: str) -> cst.SimpleString:
    """Create a SimpleString node."""
    escaped = value.replace("\\", "\\\\").replace('"', '\\"')
    return cst.SimpleString(f'"{escaped}"')


def make_docstring(text: str) -> cst.SimpleStatementLine:
    """Create a docstring statement."""
    escaped = text.replace('"""', r"\"\"\"")
    return cst.SimpleStatementLine(
        body=[cst.Expr(value=cst.SimpleString(f'"""{escaped}"""'))]
    )


def make_assign(
    target: str,
    value: cst.BaseExpression,
    annotation: str | None = None,
) -> cst.SimpleStatementLine:
    """Create an assignment statement."""
    if annotation:
        return cst.SimpleStatementLine(
            body=[
                cst.AnnAssign(
                    target=cst.Name(target),
                    annotation=make_annotation(annotation),
                    value=value,
                )
            ]
        )
    return cst.SimpleStatementLine(
        body=[
            cst.Assign(
                targets=[cst.AssignTarget(target=cst.Name(target))],
                value=value,
            )
        ]
    )


def make_import_from(
    module: str,
    names: Sequence[str],
) -> cst.SimpleStatementLine:
    """Create an import from statement."""
    import_names = [cst.ImportAlias(name=cst.Name(n)) for n in names]
    return cst.SimpleStatementLine(
        body=[
            cst.ImportFrom(
                module=cst.Attribute(
                    value=cst.parse_expression(module.rsplit(".", 1)[0]),
                    attr=cst.Name(module.rsplit(".", 1)[1]),
                )
                if "." in module
                else cst.Name(module),
                names=import_names,
            )
        ]
    )


def make_module(
    body: Sequence[cst.SimpleStatementLine | cst.BaseCompoundStatement],
) -> cst.Module:
    """Create a Module node."""
    return cst.Module(body=list(body))


def make_class_def(
    name: str,
    bases: Sequence[str],
    body: Sequence[cst.BaseStatement],
    decorators: Sequence[cst.Decorator] | None = None,
) -> cst.ClassDef:
    """Create a ClassDef node."""
    base_args = [cst.Arg(value=cst.parse_expression(b)) for b in bases]
    return cst.ClassDef(
        name=cst.Name(name),
        bases=base_args,
        body=cst.IndentedBlock(body=list(body)),
        decorators=list(decorators) if decorators else [],
    )


def make_function_def(
    name: str,
    params: Sequence[cst.Param],
    body: Sequence[cst.BaseStatement],
    returns: str | None = None,
    decorators: Sequence[cst.Decorator] | None = None,
    is_async: bool = False,
) -> cst.FunctionDef:
    """Create a FunctionDef node."""
    return_annotation = make_annotation(returns) if returns else None
    func = cst.FunctionDef(
        name=cst.Name(name),
        params=cst.Parameters(params=list(params)),
        body=cst.IndentedBlock(body=list(body)),
        returns=return_annotation,
        decorators=list(decorators) if decorators else [],
        asynchronous=cst.Asynchronous() if is_async else None,
    )
    return func


def make_param(
    name: str,
    annotation: str | None = None,
    default: cst.BaseExpression | None = None,
) -> cst.Param:
    """Create a function parameter."""
    return cst.Param(
        name=cst.Name(name),
        annotation=make_annotation(annotation) if annotation else None,
        default=default,
    )



"""Generated API client for httpbin."""

from ._base import APIError, ResponseValue
from .client import Client
from .types import (
    BearerResponse,
    HeadersResponse,
    HttpMethodResponse,
    IpResponse,
    PostResponse,
    UserAgentResponse,
)

__all__ = [
    "APIError",
    "ResponseValue",
    "Client",
    "BearerResponse",
    "HeadersResponse",
    "HttpMethodResponse",
    "IpResponse",
    "PostResponse",
    "UserAgentResponse",
]



from __future__ import annotations
from typing import final
import httpx
import msgspec
from typing import Any
from ._base import APIError, ResponseValue, encode_path
from .types import (
    BearerResponse,
    HeadersResponse,
    HttpMethodResponse,
    IpResponse,
    PostResponse,
    UserAgentResponse,
)


@final
class Client:
    def __init__(self, base_url: str) -> None:
        self._base_url = base_url
        self._client = httpx.AsyncClient(base_url=base_url)

    async def close(self) -> None:
        await self._client.aclose()

    async def __aenter__(self) -> "Client":
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: object,
    ) -> None:
        await self.close()

    async def get_request(self) -> ResponseValue[HttpMethodResponse]:
        """Returns the GET request's data"""
        _path = "/get"
        _url = self._base_url.rstrip("/") + _path
        _response = await self._client.request(
            method="GET",
            url=_url,
            headers={},
            params={},
            content=None,
        )
        _headers_dict = dict(_response.headers)
        if _response.status_code >= 400:
            raise APIError(
                status_code=_response.status_code,
                body=_response.content,
                headers=_headers_dict,
            )
        _result = msgspec.json.decode(_response.content, type=HttpMethodResponse)
        return ResponseValue(
            value=_result, status_code=_response.status_code, headers=_headers_dict
        )

    async def post_request(
        self, body: dict[str, Any] | None = None
    ) -> ResponseValue[PostResponse]:
        """Returns the POSTed data"""
        _path = "/post"
        _url = self._base_url.rstrip("/") + _path
        _response = await self._client.request(
            method="POST",
            url=_url,
            headers={"Content-Type": "application/json"},
            params={},
            content=msgspec.json.encode(body) if body is not None else None,
        )
        _headers_dict = dict(_response.headers)
        if _response.status_code >= 400:
            raise APIError(
                status_code=_response.status_code,
                body=_response.content,
                headers=_headers_dict,
            )
        _result = msgspec.json.decode(_response.content, type=PostResponse)
        return ResponseValue(
            value=_result, status_code=_response.status_code, headers=_headers_dict
        )

    async def get_origin_ip(self) -> ResponseValue[IpResponse]:
        """Returns Origin IP"""
        _path = "/ip"
        _url = self._base_url.rstrip("/") + _path
        _response = await self._client.request(
            method="GET",
            url=_url,
            headers={},
            params={},
            content=None,
        )
        _headers_dict = dict(_response.headers)
        if _response.status_code >= 400:
            raise APIError(
                status_code=_response.status_code,
                body=_response.content,
                headers=_headers_dict,
            )
        _result = msgspec.json.decode(_response.content, type=IpResponse)
        return ResponseValue(
            value=_result, status_code=_response.status_code, headers=_headers_dict
        )

    async def get_user_agent(self) -> ResponseValue[UserAgentResponse]:
        """Returns the user agent"""
        _path = "/user-agent"
        _url = self._base_url.rstrip("/") + _path
        _response = await self._client.request(
            method="GET",
            url=_url,
            headers={},
            params={},
            content=None,
        )
        _headers_dict = dict(_response.headers)
        if _response.status_code >= 400:
            raise APIError(
                status_code=_response.status_code,
                body=_response.content,
                headers=_headers_dict,
            )
        _result = msgspec.json.decode(_response.content, type=UserAgentResponse)
        return ResponseValue(
            value=_result, status_code=_response.status_code, headers=_headers_dict
        )

    async def get_headers(self) -> ResponseValue[HeadersResponse]:
        """Returns the request headers"""
        _path = "/headers"
        _url = self._base_url.rstrip("/") + _path
        _response = await self._client.request(
            method="GET",
            url=_url,
            headers={},
            params={},
            content=None,
        )
        _headers_dict = dict(_response.headers)
        if _response.status_code >= 400:
            raise APIError(
                status_code=_response.status_code,
                body=_response.content,
                headers=_headers_dict,
            )
        _result = msgspec.json.decode(_response.content, type=HeadersResponse)
        return ResponseValue(
            value=_result, status_code=_response.status_code, headers=_headers_dict
        )

    async def get_status_code(self, status_code: int) -> ResponseValue[None]:
        """Returns the specified HTTP status code"""
        _path = f"/status/{encode_path(status_code)}"
        _url = self._base_url.rstrip("/") + _path
        _response = await self._client.request(
            method="GET",
            url=_url,
            headers={},
            params={},
            content=None,
        )
        _headers_dict = dict(_response.headers)
        if _response.status_code >= 400:
            raise APIError(
                status_code=_response.status_code,
                body=_response.content,
                headers=_headers_dict,
            )
        return ResponseValue(
            value=None, status_code=_response.status_code, headers=_headers_dict
        )

    async def get_delay(self, n: int) -> ResponseValue[HttpMethodResponse]:
        """Delays responding for n seconds"""
        _path = f"/delay/{encode_path(n)}"
        _url = self._base_url.rstrip("/") + _path
        _response = await self._client.request(
            method="GET",
            url=_url,
            headers={},
            params={},
            content=None,
        )
        _headers_dict = dict(_response.headers)
        if _response.status_code >= 400:
            raise APIError(
                status_code=_response.status_code,
                body=_response.content,
                headers=_headers_dict,
            )
        _result = msgspec.json.decode(_response.content, type=HttpMethodResponse)
        return ResponseValue(
            value=_result, status_code=_response.status_code, headers=_headers_dict
        )

    async def test_bearer_auth(self) -> ResponseValue[BearerResponse]:
        """Tests Bearer authentication"""
        _path = "/bearer"
        _url = self._base_url.rstrip("/") + _path
        _response = await self._client.request(
            method="GET",
            url=_url,
            headers={},
            params={},
            content=None,
        )
        _headers_dict = dict(_response.headers)
        if _response.status_code >= 400:
            raise APIError(
                status_code=_response.status_code,
                body=_response.content,
                headers=_headers_dict,
            )
        _result = msgspec.json.decode(_response.content, type=BearerResponse)
        return ResponseValue(
            value=_result, status_code=_response.status_code, headers=_headers_dict
        )



"""Pytest configuration and fixtures."""

from __future__ import annotations

import json
import subprocess
from pathlib import Path
from typing import Any

import pytest
from dotenv import load_dotenv
from inline_snapshot import register_format
from inline_snapshot._external._diff import TextDiff

# Load .env file at test startup (handles multi-line values properly)
load_dotenv()


def format_python(code: str) -> str:
    """Format Python code with Ruff."""
    result = subprocess.run(
        ["uv", "run", "ruff", "check", "--select=F401", "--fix", "-"],
        input=code,
        capture_output=True,
        text=True,
        check=False,
    )
    code = result.stdout or code
    result = subprocess.run(
        ["uv", "run", "ruff", "format", "-"],
        input=code,
        capture_output=True,
        text=True,
        check=True,
    )
    return result.stdout


@register_format
class PythonFormat(TextDiff):
    """Format handler for .py files with Ruff formatting."""

    suffix = ".py"
    priority = 0

    def is_format_for(self, value: object) -> bool:
        return isinstance(value, str)

    def encode(self, value: str, path: Path) -> None:
        formatted = format_python(value)
        with path.open("w", encoding="utf-8", newline="\n") as f:
            f.write(formatted)

    def decode(self, path: Path) -> str:
        with path.open("r", encoding="utf-8", newline="\n") as f:
            return f.read()


@pytest.fixture
def fixtures_dir() -> Path:
    """Return the fixtures directory."""
    return Path(__file__).parent / "fixtures"


@pytest.fixture
def httpbin_spec(fixtures_dir: Path) -> dict[str, Any]:
    """Load the httpbin spec."""
    with open(fixtures_dir / "httpbin.json") as f:
        return json.load(f)  # type: ignore[no-any-return]


@pytest.fixture
def allof_spec(fixtures_dir: Path) -> dict[str, Any]:
    """Load the allOf test spec."""
    with open(fixtures_dir / "allof_test.json") as f:
        return json.load(f)  # type: ignore[no-any-return]


@pytest.fixture
def edge_cases_spec(fixtures_dir: Path) -> dict[str, Any]:
    """Load the edge cases test spec."""
    with open(fixtures_dir / "edge_cases.json") as f:
        return json.load(f)  # type: ignore[no-any-return]


@pytest.fixture
def kalshi_spec(fixtures_dir: Path) -> dict[str, Any]:
    """Load the Kalshi OpenAPI spec."""
    import yaml

    with open(fixtures_dir / "kalshi.yaml") as f:
        return yaml.safe_load(f)  # type: ignore[no-any-return]


@pytest.fixture
def pagination_spec(fixtures_dir: Path) -> dict[str, Any]:
    """Load the pagination test spec."""
    import yaml

    with open(fixtures_dir / "pagination.yaml") as f:
        return yaml.safe_load(f)  # type: ignore[no-any-return]



from __future__ import annotations
from typing import final
import httpx
import msgspec
from typing import Any
from ._base import APIError, ResponseValue, SyncPaginator
from .types import GetItemsResponse, Item, ListOrdersResponse, ListUsersResponse, User


@final
class Client:
    def __init__(self, base_url: str) -> None:
        self._base_url = base_url
        self._client = httpx.Client(base_url=base_url)

    def close(self) -> None:
        self._client.close()

    def __enter__(self) -> "Client":
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: object,
    ) -> None:
        self.close()

    def list_users(
        self, limit: int | None = None, cursor: str | None = None
    ) -> ResponseValue[ListUsersResponse]:
        """List all users

        Paginated list of users"""
        _path = "/users"
        _params: dict[str, Any] = {"limit": limit, "cursor": cursor}
        _url = self._base_url.rstrip("/") + _path
        _response = self._client.request(
            method="GET",
            url=_url,
            headers={},
            params={k: v for k, v in _params.items() if v is not None},
            content=None,
        )
        _headers_dict = dict(_response.headers)
        if _response.status_code >= 400:
            raise APIError(
                status_code=_response.status_code,
                body=_response.content,
                headers=_headers_dict,
            )
        _result = msgspec.json.decode(_response.content, type=ListUsersResponse)
        return ResponseValue(
            value=_result, status_code=_response.status_code, headers=_headers_dict
        )

    def list_users_iter(self, limit: int | None = None) -> SyncPaginator[User]:
        return SyncPaginator(
            fetch_page=lambda _token: self.list_users(limit=limit, cursor=_token),
            get_items=lambda page: list(page.items),
            get_next_token=lambda page: page.cursor or None,
        )

    def get_items(
        self, per_page: int | None = None, page_token: str | None = None
    ) -> ResponseValue[GetItemsResponse]:
        """Get items with offset pagination

        Items list with different pagination fields"""
        _path = "/items"
        _params: dict[str, Any] = {"per_page": per_page, "page_token": page_token}
        _url = self._base_url.rstrip("/") + _path
        _response = self._client.request(
            method="GET",
            url=_url,
            headers={},
            params={k: v for k, v in _params.items() if v is not None},
            content=None,
        )
        _headers_dict = dict(_response.headers)
        if _response.status_code >= 400:
            raise APIError(
                status_code=_response.status_code,
                body=_response.content,
                headers=_headers_dict,
            )
        _result = msgspec.json.decode(_response.content, type=GetItemsResponse)
        return ResponseValue(
            value=_result, status_code=_response.status_code, headers=_headers_dict
        )

    def get_items_iter(self, per_page: int | None = None) -> SyncPaginator[Item]:
        return SyncPaginator(
            fetch_page=lambda _token: self.get_items(
                per_page=per_page, page_token=_token
            ),
            get_items=lambda page: list(page.data),
            get_next_token=lambda page: page.next_cursor or None,
        )

    def list_orders(
        self, status: str | None = None
    ) -> ResponseValue[ListOrdersResponse]:
        """List orders (not paginated - no config)"""
        _path = "/orders"
        _params: dict[str, Any] = {"status": status}
        _url = self._base_url.rstrip("/") + _path
        _response = self._client.request(
            method="GET",
            url=_url,
            headers={},
            params={k: v for k, v in _params.items() if v is not None},
            content=None,
        )
        _headers_dict = dict(_response.headers)
        if _response.status_code >= 400:
            raise APIError(
                status_code=_response.status_code,
                body=_response.content,
                headers=_headers_dict,
            )
        _result = msgspec.json.decode(_response.content, type=ListOrdersResponse)
        return ResponseValue(
            value=_result, status_code=_response.status_code, headers=_headers_dict
        )



from .models import (
    BearerResponse,
    HeadersResponse,
    HttpMethodResponse,
    IpResponse,
    PostResponse,
    UserAgentResponse,
)

__all__ = [
    "BearerResponse",
    "HeadersResponse",
    "HttpMethodResponse",
    "IpResponse",
    "PostResponse",
    "UserAgentResponse",
]



"""Minimal unit tests for behaviors hard to express in fixtures."""

from __future__ import annotations

import pytest

from senzo.type_space import to_pascal_case, to_snake_case


@pytest.mark.parametrize(
    "input,expected",
    [
        ("CamelCase", "camel_case"),
        ("camelCase", "camel_case"),
        ("snake_case", "snake_case"),
        ("HTTPRequest", "http_request"),
        ("XMLParser", "xml_parser"),
        ("IOError", "io_error"),
        ("createdAt", "created_at"),
        ("userID", "user_id"),
    ],
)
def test_to_snake_case(input: str, expected: str) -> None:
    assert to_snake_case(input) == expected


@pytest.mark.parametrize(
    "input,expected",
    [
        ("camel_case", "CamelCase"),
        ("http_request", "HttpRequest"),
        ("Pet", "Pet"),
        ("CreatePetRequest", "CreatePetRequest"),
    ],
)
def test_to_pascal_case(input: str, expected: str) -> None:
    assert to_pascal_case(input) == expected



"""httpx backend for HTTP client generation."""

from __future__ import annotations

import libcst as cst

from senzo.backends.base import HttpBackend
from senzo.codegen.utils import make_annotation, make_docstring, make_param
from senzo.operation import (
    ContentType,
    OperationMethod,
    OperationParameter,
    OperationResponse,
    ParameterLocation,
)


class HttpxBackend(HttpBackend):
    """httpx backend - supports both sync and async.

    Generates fully functional API client methods with proper request building,
    response parsing, and error handling.
    """

    def __init__(self, async_mode: bool = False) -> None:
        super().__init__(async_mode=async_mode)

    def generate_client_class(
        self,
        name: str,
        base_url_param: bool = True,
    ) -> cst.ClassDef:
        """Generate the client class with constructor."""
        init_body: list[cst.BaseStatement] = []

        # self._base_url = base_url
        init_body.append(
            cst.SimpleStatementLine(
                body=[
                    cst.Assign(
                        targets=[
                            cst.AssignTarget(
                                target=cst.Attribute(
                                    value=cst.Name("self"),
                                    attr=cst.Name("_base_url"),
                                )
                            )
                        ],
                        value=cst.Name("base_url"),
                    )
                ]
            )
        )

        # self._client = httpx.Client() or httpx.AsyncClient()
        client_class = "AsyncClient" if self.async_mode else "Client"
        init_body.append(
            cst.SimpleStatementLine(
                body=[
                    cst.Assign(
                        targets=[
                            cst.AssignTarget(
                                target=cst.Attribute(
                                    value=cst.Name("self"),
                                    attr=cst.Name("_client"),
                                )
                            )
                        ],
                        value=cst.Call(
                            func=cst.Attribute(
                                value=cst.Name("httpx"),
                                attr=cst.Name(client_class),
                            ),
                            args=[
                                cst.Arg(
                                    keyword=cst.Name("base_url"),
                                    value=cst.Name("base_url"),
                                )
                            ],
                        ),
                    )
                ]
            )
        )

        # Build init params - add inner if hooks configured
        init_params = [
            cst.Param(name=cst.Name("self")),
            cst.Param(
                name=cst.Name("base_url"),
                annotation=cst.Annotation(annotation=cst.Name("str")),
            ),
        ]

        # Add inner parameter if hooks are configured
        if self._hooks_config is not None:
            inner_class = self._hooks_config.get_inner_class()
            init_params.append(
                cst.Param(
                    name=cst.Name("inner"),
                    annotation=cst.Annotation(annotation=cst.Name(inner_class)),
                )
            )
            # self._inner = inner
            init_body.append(
                cst.SimpleStatementLine(
                    body=[
                        cst.Assign(
                            targets=[
                                cst.AssignTarget(
                                    target=cst.Attribute(
                                        value=cst.Name("self"),
                                        attr=cst.Name("_inner"),
                                    )
                                )
                            ],
                            value=cst.Name("inner"),
                        )
                    ]
                )
            )

        init_func = cst.FunctionDef(
            name=cst.Name("__init__"),
            params=cst.Parameters(params=init_params),
            body=cst.IndentedBlock(body=init_body),
            returns=cst.Annotation(annotation=cst.Name("None")),
        )

        # Generate close method
        close_body: list[cst.BaseStatement] = [
            cst.SimpleStatementLine(
                body=[
                    cst.Expr(
                        value=cst.Call(
                            func=cst.Attribute(
                                value=cst.Attribute(
                                    value=cst.Name("self"),
                                    attr=cst.Name("_client"),
                                ),
                                attr=cst.Name("close"),
                            ),
                            args=[],
                        )
                        if not self.async_mode
                        else cst.Await(
                            expression=cst.Call(
                                func=cst.Attribute(
                                    value=cst.Attribute(
                                        value=cst.Name("self"),
                                        attr=cst.Name("_client"),
                                    ),
                                    attr=cst.Name("aclose"),
                                ),
                                args=[],
                            )
                        )
                    )
                ]
            )
        ]

        close_func = cst.FunctionDef(
            name=cst.Name("close"),
            params=cst.Parameters(params=[cst.Param(name=cst.Name("self"))]),
            body=cst.IndentedBlock(body=close_body),
            returns=cst.Annotation(annotation=cst.Name("None")),
            asynchronous=cst.Asynchronous() if self.async_mode else None,
        )

        # Context manager methods
        enter_body = [
            cst.SimpleStatementLine(body=[cst.Return(value=cst.Name("self"))])
        ]
        exit_body = [
            cst.SimpleStatementLine(
                body=[
                    cst.Expr(
                        value=cst.Call(
                            func=cst.Attribute(
                                value=cst.Name("self"),
                                attr=cst.Name("close"),
                            ),
                            args=[],
                        )
                        if not self.async_mode
                        else cst.Await(
                            expression=cst.Call(
                                func=cst.Attribute(
                                    value=cst.Name("self"),
                                    attr=cst.Name("close"),
                                ),
                                args=[],
                            )
                        )
                    )
                ]
            )
        ]

        enter_name = "__aenter__" if self.async_mode else "__enter__"
        exit_name = "__aexit__" if self.async_mode else "__exit__"
        self_type = f'"{name}"'

        enter_func = cst.FunctionDef(
            name=cst.Name(enter_name),
            params=cst.Parameters(params=[cst.Param(name=cst.Name("self"))]),
            body=cst.IndentedBlock(body=enter_body),
            returns=make_annotation(self_type),
            asynchronous=cst.Asynchronous() if self.async_mode else None,
        )

        exit_params = [
            cst.Param(name=cst.Name("self")),
            cst.Param(
                name=cst.Name("exc_type"),
                annotation=make_annotation("type[BaseException] | None"),
            ),
            cst.Param(
                name=cst.Name("exc_val"),
                annotation=make_annotation("BaseException | None"),
            ),
            cst.Param(
                name=cst.Name("exc_tb"),
                annotation=make_annotation("object"),
            ),
        ]

        exit_func = cst.FunctionDef(
            name=cst.Name(exit_name),
            params=cst.Parameters(params=exit_params),
            body=cst.IndentedBlock(body=exit_body),
            returns=cst.Annotation(annotation=cst.Name("None")),
            asynchronous=cst.Asynchronous() if self.async_mode else None,
        )

        class_body: list[cst.BaseStatement] = [
            init_func,
            close_func,
            enter_func,
            exit_func,
        ]

        return cst.ClassDef(
            name=cst.Name(name),
            body=cst.IndentedBlock(body=class_body),
            decorators=[cst.Decorator(decorator=cst.Name("final"))],
        )

    def generate_sub_client_class(
        self,
        name: str,
    ) -> cst.ClassDef:
        """Generate a sub-client class that shares an existing session.

        Sub-clients accept an existing HTTP client instead of creating their own.
        They have no lifecycle methods (close, context manager) - the parent owns the session.

        Sub-clients still need _base_url for building full URLs in hooks.
        """
        init_body: list[cst.BaseStatement] = []

        # self._base_url = base_url
        init_body.append(
            cst.SimpleStatementLine(
                body=[
                    cst.Assign(
                        targets=[
                            cst.AssignTarget(
                                target=cst.Attribute(
                                    value=cst.Name("self"),
                                    attr=cst.Name("_base_url"),
                                )
                            )
                        ],
                        value=cst.Name("base_url"),
                    )
                ]
            )
        )

        # self._client = client
        init_body.append(
            cst.SimpleStatementLine(
                body=[
                    cst.Assign(
                        targets=[
                            cst.AssignTarget(
                                target=cst.Attribute(
                                    value=cst.Name("self"),
                                    attr=cst.Name("_client"),
                                )
                            )
                        ],
                        value=cst.Name("client"),
                    )
                ]
            )
        )

        # Build init params
        client_type = "httpx.AsyncClient" if self.async_mode else "httpx.Client"
        init_params = [
            cst.Param(name=cst.Name("self")),
            cst.Param(
                name=cst.Name("client"),
                annotation=make_annotation(client_type),
            ),
            cst.Param(
                name=cst.Name("base_url"),
                annotation=cst.Annotation(annotation=cst.Name("str")),
            ),
        ]

        # Add inner parameter if hooks are configured
        if self._hooks_config is not None:
            inner_class = self._hooks_config.get_inner_class()
            init_params.append(
                cst.Param(
                    name=cst.Name("inner"),
                    annotation=cst.Annotation(annotation=cst.Name(inner_class)),
                )
            )
            # self._inner = inner
            init_body.append(
                cst.SimpleStatementLine(
                    body=[
                        cst.Assign(
                            targets=[
                                cst.AssignTarget(
                                    target=cst.Attribute(
                                        value=cst.Name("self"),
                                        attr=cst.Name("_inner"),
                                    )
                                )
                            ],
                            value=cst.Name("inner"),
                        )
                    ]
                )
            )

        init_func = cst.FunctionDef(
            name=cst.Name("__init__"),
            params=cst.Parameters(params=init_params),
            body=cst.IndentedBlock(body=init_body),
            returns=cst.Annotation(annotation=cst.Name("None")),
        )

        class_body: list[cst.BaseStatement] = [init_func]

        return cst.ClassDef(
            name=cst.Name(name),
            body=cst.IndentedBlock(body=class_body),
            decorators=[cst.Decorator(decorator=cst.Name("final"))],
        )

    def generate_method(
        self,
        operation: OperationMethod,
    ) -> cst.FunctionDef:
        """Generate an API method with full request/response handling."""
        method_name = operation.python_method_name()
        params = self._build_params(operation)
        body = self._build_method_body(operation)
        return_type = self._get_return_type(operation)

        decorators: list[cst.Decorator] = []
        if operation.deprecated:
            decorators.append(
                cst.Decorator(
                    decorator=cst.Call(
                        func=cst.Name("deprecated"),
                        args=[
                            cst.Arg(
                                value=cst.SimpleString(
                                    f'"{operation.operation_id} is deprecated"'
                                )
                            )
                        ],
                    )
                )
            )

        func = cst.FunctionDef(
            name=cst.Name(method_name),
            params=cst.Parameters(params=params),
            body=cst.IndentedBlock(body=body),
            returns=make_annotation(return_type),
            asynchronous=cst.Asynchronous() if self.async_mode else None,
            decorators=decorators if decorators else [],
        )

        # Add docstring if summary or description exists
        if operation.summary or operation.description:
            doc_text = operation.summary or ""
            if operation.description and operation.description != operation.summary:
                if doc_text:
                    doc_text += "\n\n"
                doc_text += operation.description
            doc_stmt = make_docstring(doc_text)
            new_body = [doc_stmt, *body]
            func = func.with_changes(body=cst.IndentedBlock(body=new_body))

        return func

    def _build_params(
        self,
        operation: OperationMethod,
    ) -> list[cst.Param]:
        """Build function parameters from operation parameters."""
        params: list[cst.Param] = [cst.Param(name=cst.Name("self"))]

        # Add required parameters first
        for param in operation.parameters:
            if param.required:
                params.append(
                    make_param(param.name, param.type_annotation, default=None)
                )

        # Add request body if required
        if operation.request_body and operation.request_body.required:
            params.append(
                make_param("body", operation.request_body.type_annotation, default=None)
            )

        # Add optional parameters
        for param in operation.parameters:
            if not param.required:
                params.append(
                    make_param(
                        param.name,
                        param.type_annotation,
                        default=cst.Name("None"),
                    )
                )

        # Add optional request body
        if operation.request_body and not operation.request_body.required:
            params.append(
                make_param(
                    "body",
                    f"{operation.request_body.type_annotation} | None",
                    default=cst.Name("None"),
                )
            )

        return params

    def _build_method_body(
        self,
        operation: OperationMethod,
    ) -> list[cst.BaseStatement]:
        """Build the method body with request construction and response handling."""
        statements: list[cst.BaseStatement] = []

        # Build OperationInfo only if hooks are configured
        if self._hooks_config is not None:
            statements.append(self._build_operation_info(operation))

        # Build URL path
        path_expr = self._build_path_expression(operation)
        statements.append(
            cst.SimpleStatementLine(
                body=[
                    cst.Assign(
                        targets=[cst.AssignTarget(target=cst.Name("_path"))],
                        value=path_expr,
                    )
                ]
            )
        )

        # Build query params
        query_params = [
            p for p in operation.parameters if p.location == ParameterLocation.QUERY
        ]
        if query_params:
            statements.append(self._build_query_params(query_params))

        # Build headers
        header_params = [
            p for p in operation.parameters if p.location == ParameterLocation.HEADER
        ]
        if header_params:
            statements.append(self._build_header_params(header_params))

        # Build request execution
        if self._hooks_config is not None:
            # With hooks: build RequestData, call hook functions
            statements.append(
                self._build_request_data(operation, query_params, header_params)
            )
            statements.extend(self._build_hooked_execution(operation))
        else:
            # Without hooks: direct HTTP call
            statements.extend(
                self._build_direct_execution(operation, query_params, header_params)
            )

        # Handle response
        statements.extend(self._build_response_handling(operation))

        return statements

    def _build_operation_info(
        self,
        operation: OperationMethod,
    ) -> cst.SimpleStatementLine:
        """Build OperationInfo dataclass instantiation."""
        tags_list = cst.List(
            elements=[
                cst.Element(cst.SimpleString(f'"{tag}"')) for tag in operation.tags
            ]
        )
        security_list = cst.List(
            elements=[
                cst.Element(cst.SimpleString(f'"{scheme}"'))
                for scheme in operation.security
            ]
        )

        return cst.SimpleStatementLine(
            body=[
                cst.Assign(
                    targets=[cst.AssignTarget(target=cst.Name("_info"))],
                    value=cst.Call(
                        func=cst.Name("OperationInfo"),
                        args=[
                            cst.Arg(
                                keyword=cst.Name("operation_id"),
                                value=cst.SimpleString(f'"{operation.operation_id}"'),
                            ),
                            cst.Arg(
                                keyword=cst.Name("method"),
                                value=cst.SimpleString(
                                    f'"{operation.method.value.upper()}"'
                                ),
                            ),
                            cst.Arg(
                                keyword=cst.Name("path_template"),
                                value=cst.SimpleString(f'"{operation.path.template}"'),
                            ),
                            cst.Arg(
                                keyword=cst.Name("tags"),
                                value=tags_list,
                            ),
                            cst.Arg(
                                keyword=cst.Name("security"),
                                value=security_list,
                            ),
                        ],
                    ),
                )
            ]
        )

    def _build_request_data(
        self,
        operation: OperationMethod,
        query_params: list[OperationParameter],
        header_params: list[OperationParameter],
    ) -> cst.SimpleStatementLine:
        """Build RequestData dataclass instantiation."""
        # URL: self._base_url.rstrip("/") + _path
        # This handles both base URLs with and without trailing slashes
        url_expr = cst.BinaryOperation(
            left=cst.Call(
                func=cst.Attribute(
                    value=cst.Attribute(
                        value=cst.Name("self"),
                        attr=cst.Name("_base_url"),
                    ),
                    attr=cst.Name("rstrip"),
                ),
                args=[cst.Arg(value=cst.SimpleString('"/"'))],
            ),
            operator=cst.Add(),
            right=cst.Name("_path"),
        )

        # Headers dict
        if header_params:
            headers_expr = cst.Name("_headers")
        else:
            headers_expr = cst.Dict(elements=[])

        # Add Content-Type header for JSON body
        if (
            operation.request_body
            and operation.request_body.content_type == ContentType.JSON
        ):
            if header_params:
                # Merge with existing headers
                headers_expr = cst.DictComp(
                    key=cst.Name("k"),
                    value=cst.Name("v"),
                    for_in=cst.CompFor(
                        target=cst.Tuple(
                            elements=[
                                cst.Element(cst.Name("k")),
                                cst.Element(cst.Name("v")),
                            ]
                        ),
                        iter=cst.Call(
                            func=cst.Attribute(
                                value=cst.BinaryOperation(
                                    left=cst.Name("_headers"),
                                    operator=cst.BitOr(),
                                    right=cst.Dict(
                                        elements=[
                                            cst.DictElement(
                                                key=cst.SimpleString('"Content-Type"'),
                                                value=cst.SimpleString(
                                                    '"application/json"'
                                                ),
                                            )
                                        ]
                                    ),
                                ),
                                attr=cst.Name("items"),
                            ),
                            args=[],
                        ),
                        ifs=[
                            cst.CompIf(
                                test=cst.Comparison(
                                    left=cst.Name("v"),
                                    comparisons=[
                                        cst.ComparisonTarget(
                                            operator=cst.IsNot(),
                                            comparator=cst.Name("None"),
                                        )
                                    ],
                                )
                            )
                        ],
                    ),
                )
            else:
                headers_expr = cst.Dict(
                    elements=[
                        cst.DictElement(
                            key=cst.SimpleString('"Content-Type"'),
                            value=cst.SimpleString('"application/json"'),
                        )
                    ]
                )
        elif header_params:
            # Filter None values from headers
            headers_expr = cst.DictComp(
                key=cst.Name("k"),
                value=cst.Name("v"),
                for_in=cst.CompFor(
                    target=cst.Tuple(
                        elements=[
                            cst.Element(cst.Name("k")),
                            cst.Element(cst.Name("v")),
                        ]
                    ),
                    iter=cst.Call(
                        func=cst.Attribute(
                            value=cst.Name("_headers"),
                            attr=cst.Name("items"),
                        ),
                        args=[],
                    ),
                    ifs=[
                        cst.CompIf(
                            test=cst.Comparison(
                                left=cst.Name("v"),
                                comparisons=[
                                    cst.ComparisonTarget(
                                        operator=cst.IsNot(),
                                        comparator=cst.Name("None"),
                                    )
                                ],
                            )
                        )
                    ],
                ),
            )

        # Params dict
        if query_params:
            params_expr = cst.Name("_params")
        else:
            params_expr = cst.Dict(elements=[])

        # Content (body)
        content_expr: cst.BaseExpression = cst.Name("None")
        if operation.request_body:
            if operation.request_body.content_type == ContentType.JSON:
                if self._dataclass_backend is not None:
                    encode_expr = (
                        self._dataclass_backend.generate_encode_json_expression(
                            operation.request_body.type_annotation or "Any",
                            cst.Name("body"),
                        )
                    )
                    if operation.request_body.required:
                        content_expr = encode_expr
                    else:
                        content_expr = cst.IfExp(
                            test=cst.Comparison(
                                left=cst.Name("body"),
                                comparisons=[
                                    cst.ComparisonTarget(
                                        operator=cst.IsNot(),
                                        comparator=cst.Name("None"),
                                    )
                                ],
                            ),
                            body=encode_expr,
                            orelse=cst.Name("None"),
                        )

        return cst.SimpleStatementLine(
            body=[
                cst.Assign(
                    targets=[cst.AssignTarget(target=cst.Name("_request"))],
                    value=cst.Call(
                        func=cst.Name("RequestData"),
                        args=[
                            cst.Arg(
                                keyword=cst.Name("method"),
                                value=cst.SimpleString(
                                    f'"{operation.method.value.upper()}"'
                                ),
                            ),
                            cst.Arg(
                                keyword=cst.Name("url"),
                                value=url_expr,
                            ),
                            cst.Arg(
                                keyword=cst.Name("headers"),
                                value=headers_expr,
                            ),
                            cst.Arg(
                                keyword=cst.Name("params"),
                                value=params_expr,
                            ),
                            cst.Arg(
                                keyword=cst.Name("content"),
                                value=content_expr,
                            ),
                        ],
                    ),
                )
            ]
        )

    def _build_hooked_execution(
        self,
        operation: OperationMethod,
    ) -> list[cst.BaseStatement]:
        """Build execution with hook function calls baked in."""
        assert self._hooks_config is not None

        statements: list[cst.BaseStatement] = []
        await_prefix = "await " if self.async_mode else ""

        # Pre-hook: _request = pre_hook(self._inner, _request, _info)
        if self._hooks_config.pre_hook:
            assert isinstance(self._hooks_config.pre_hook, str)
            pre_hook_func = self._hooks_config.get_hook_function(
                self._hooks_config.pre_hook
            )
            code = f"_request = {await_prefix}{pre_hook_func}(self._inner, _request, _info)"
            statements.append(cst.parse_statement(code))

        # Execute HTTP request
        if self.async_mode:
            exec_code = """
_resp = await self._client.request(
    method=_request.method,
    url=_request.url,
    headers=_request.headers,
    params={k: v for k, v in _request.params.items() if v is not None},
    content=_request.content,
)
_response = ResponseData(
    status_code=_resp.status_code,
    headers=dict(_resp.headers),
    content=_resp.content,
)
"""
        else:
            exec_code = """
_resp = self._client.request(
    method=_request.method,
    url=_request.url,
    headers=_request.headers,
    params={k: v for k, v in _request.params.items() if v is not None},
    content=_request.content,
)
_response = ResponseData(
    status_code=_resp.status_code,
    headers=dict(_resp.headers),
    content=_resp.content,
)
"""
        exec_module = cst.parse_module(exec_code.strip())
        statements.extend(exec_module.body)  # type: ignore[arg-type]

        # Post-hook: _response = post_hook(self._inner, _response, _info)
        if self._hooks_config.post_hook:
            assert isinstance(self._hooks_config.post_hook, str)
            post_hook_func = self._hooks_config.get_hook_function(
                self._hooks_config.post_hook
            )
            code = f"_response = {await_prefix}{post_hook_func}(self._inner, _response, _info)"
            statements.append(cst.parse_statement(code))

        return statements

    def _build_direct_execution(
        self,
        operation: OperationMethod,
        query_params: list[OperationParameter],
        header_params: list[OperationParameter],
    ) -> list[cst.BaseStatement]:
        """Build direct HTTP execution without hooks."""
        statements: list[cst.BaseStatement] = []

        # Build URL: self._base_url.rstrip("/") + _path
        url_code = '_url = self._base_url.rstrip("/") + _path'
        statements.append(cst.parse_statement(url_code))

        # Build headers
        if header_params:
            if (
                operation.request_body
                and operation.request_body.content_type == ContentType.JSON
            ):
                headers_code = '_headers_final = {k: v for k, v in ({**_headers, "Content-Type": "application/json"}).items() if v is not None}'
            else:
                headers_code = "_headers_final = {k: v for k, v in _headers.items() if v is not None}"
            statements.append(cst.parse_statement(headers_code))
            headers_var = "_headers_final"
        else:
            if (
                operation.request_body
                and operation.request_body.content_type == ContentType.JSON
            ):
                headers_var = '{"Content-Type": "application/json"}'
            else:
                headers_var = "{}"

        # Build query params filter
        if query_params:
            params_expr = "{k: v for k, v in _params.items() if v is not None}"
        else:
            params_expr = "{}"

        # Build content
        content_expr = "None"
        if operation.request_body:
            if operation.request_body.content_type == ContentType.JSON:
                if self._dataclass_backend is not None:
                    # Get the encode expression as code
                    encode_expr = (
                        self._dataclass_backend.generate_encode_json_expression(
                            operation.request_body.type_annotation or "Any",
                            cst.Name("body"),
                        )
                    )
                    encode_code = cst.Module(body=[]).code_for_node(encode_expr)
                    if operation.request_body.required:
                        content_expr = encode_code
                    else:
                        content_expr = f"{encode_code} if body is not None else None"

        # Build request call
        await_prefix = "await " if self.async_mode else ""
        request_code = f'''
_response = {await_prefix}self._client.request(
    method="{operation.method.value.upper()}",
    url=_url,
    headers={headers_var},
    params={params_expr},
    content={content_expr},
)
'''
        exec_module = cst.parse_module(request_code.strip())
        statements.extend(exec_module.body)  # type: ignore[arg-type]

        return statements

    def _build_path_expression(
        self,
        operation: OperationMethod,
    ) -> cst.BaseExpression:
        """Build the path expression with parameter substitution."""
        path_params = [
            p for p in operation.parameters if p.location == ParameterLocation.PATH
        ]

        if not path_params:
            return cst.SimpleString(f'"{operation.path.template}"')

        # Build f-string with _encode_path calls
        fstring_parts: list[cst.BaseFormattedStringContent] = []
        template = operation.path.template
        last_end = 0

        import re

        for match in re.finditer(r"\{([^}]+)\}", template):
            # Add text before parameter
            if match.start() > last_end:
                fstring_parts.append(
                    cst.FormattedStringText(value=template[last_end : match.start()])
                )

            # Find the parameter
            param_api_name = match.group(1)
            python_name = None
            for p in path_params:
                if p.api_name == param_api_name:
                    python_name = p.name
                    break
            python_name = python_name or param_api_name

            # Add formatted expression with _encode_path
            fstring_parts.append(
                cst.FormattedStringExpression(
                    expression=cst.Call(
                        func=cst.Name("encode_path"),
                        args=[cst.Arg(value=cst.Name(python_name))],
                    )
                )
            )
            last_end = match.end()

        # Add remaining text
        if last_end < len(template):
            fstring_parts.append(cst.FormattedStringText(value=template[last_end:]))

        return cst.FormattedString(parts=fstring_parts)

    def _build_query_params(
        self,
        params: list[OperationParameter],
    ) -> cst.SimpleStatementLine:
        """Build query parameters dict."""
        dict_elements: list[cst.DictElement] = []
        for param in params:
            dict_elements.append(
                cst.DictElement(
                    key=cst.SimpleString(f'"{param.api_name}"'),
                    value=cst.Name(param.name),
                )
            )

        return cst.SimpleStatementLine(
            body=[
                cst.AnnAssign(
                    target=cst.Name("_params"),
                    annotation=cst.Annotation(
                        annotation=cst.Subscript(
                            value=cst.Name("dict"),
                            slice=[
                                cst.SubscriptElement(
                                    slice=cst.Index(value=cst.Name("str"))
                                ),
                                cst.SubscriptElement(
                                    slice=cst.Index(value=cst.Name("Any"))
                                ),
                            ],
                        )
                    ),
                    value=cst.Dict(elements=dict_elements),
                )
            ]
        )

    def _build_header_params(
        self,
        params: list[OperationParameter],
    ) -> cst.SimpleStatementLine:
        """Build headers dict."""
        dict_elements: list[cst.DictElement] = []
        for param in params:
            dict_elements.append(
                cst.DictElement(
                    key=cst.SimpleString(f'"{param.api_name}"'),
                    value=cst.Name(param.name),
                )
            )

        return cst.SimpleStatementLine(
            body=[
                cst.Assign(
                    targets=[cst.AssignTarget(target=cst.Name("_headers"))],
                    value=cst.Dict(elements=dict_elements),
                )
            ]
        )

    def _build_request_call(
        self,
        operation: OperationMethod,
        query_params: list[OperationParameter],
        header_params: list[OperationParameter],
    ) -> cst.SimpleStatementLine:
        """Build the HTTP request call."""
        method = operation.method.value.lower()
        args: list[cst.Arg] = [cst.Arg(value=cst.Name("_path"))]

        if query_params:
            # Filter out None values
            filter_expr = cst.DictComp(
                key=cst.Name("k"),
                value=cst.Name("v"),
                for_in=cst.CompFor(
                    target=cst.Tuple(
                        elements=[
                            cst.Element(cst.Name("k")),
                            cst.Element(cst.Name("v")),
                        ]
                    ),
                    iter=cst.Call(
                        func=cst.Attribute(
                            value=cst.Name("_params"),
                            attr=cst.Name("items"),
                        ),
                        args=[],
                    ),
                    ifs=[
                        cst.CompIf(
                            test=cst.Comparison(
                                left=cst.Name("v"),
                                comparisons=[
                                    cst.ComparisonTarget(
                                        operator=cst.IsNot(),
                                        comparator=cst.Name("None"),
                                    )
                                ],
                            )
                        )
                    ],
                ),
            )
            args.append(cst.Arg(keyword=cst.Name("params"), value=filter_expr))

        if header_params:
            filter_expr = cst.DictComp(
                key=cst.Name("k"),
                value=cst.Name("v"),
                for_in=cst.CompFor(
                    target=cst.Tuple(
                        elements=[
                            cst.Element(cst.Name("k")),
                            cst.Element(cst.Name("v")),
                        ]
                    ),
                    iter=cst.Call(
                        func=cst.Attribute(
                            value=cst.Name("_headers"),
                            attr=cst.Name("items"),
                        ),
                        args=[],
                    ),
                    ifs=[
                        cst.CompIf(
                            test=cst.Comparison(
                                left=cst.Name("v"),
                                comparisons=[
                                    cst.ComparisonTarget(
                                        operator=cst.IsNot(),
                                        comparator=cst.Name("None"),
                                    )
                                ],
                            )
                        )
                    ],
                ),
            )
            args.append(cst.Arg(keyword=cst.Name("headers"), value=filter_expr))

        # Add request body
        if operation.request_body:
            content_type = operation.request_body.content_type
            if content_type == ContentType.JSON:
                # Use content= with bytes for JSON bodies via dataclass backend encoding
                if self._dataclass_backend is not None:
                    # Use dataclass backend's encode method
                    encode_expr = (
                        self._dataclass_backend.generate_encode_json_expression(
                            operation.request_body.type_annotation or "Any",
                            cst.Name("body"),
                        )
                    )
                    if operation.request_body.required:
                        body_arg = encode_expr
                    else:
                        # body is not None ? encode(body) : None
                        body_arg = cst.IfExp(
                            test=cst.Comparison(
                                left=cst.Name("body"),
                                comparisons=[
                                    cst.ComparisonTarget(
                                        operator=cst.IsNot(),
                                        comparator=cst.Name("None"),
                                    )
                                ],
                            ),
                            body=encode_expr,
                            orelse=cst.Name("None"),
                        )

                    args.append(cst.Arg(keyword=cst.Name("content"), value=body_arg))
                    # Add Content-Type header for bytes content
                    args.append(
                        cst.Arg(
                            keyword=cst.Name("headers"),
                            value=cst.Dict(
                                elements=[
                                    cst.DictElement(
                                        key=cst.SimpleString('"Content-Type"'),
                                        value=cst.SimpleString('"application/json"'),
                                    )
                                ]
                            ),
                        )
                    )
                else:
                    # Fallback: use json= with dict conversion (legacy behavior)
                    args.append(
                        cst.Arg(
                            keyword=cst.Name("json"),
                            value=cst.Name("body")
                            if operation.request_body.required
                            else cst.IfExp(
                                test=cst.Comparison(
                                    left=cst.Name("body"),
                                    comparisons=[
                                        cst.ComparisonTarget(
                                            operator=cst.IsNot(),
                                            comparator=cst.Name("None"),
                                        )
                                    ],
                                ),
                                body=cst.Name("body"),
                                orelse=cst.Name("None"),
                            ),
                        )
                    )

        request_call = cst.Call(
            func=cst.Attribute(
                value=cst.Attribute(
                    value=cst.Name("self"),
                    attr=cst.Name("_client"),
                ),
                attr=cst.Name(method),
            ),
            args=args,
        )

        if self.async_mode:
            request_call = cst.Await(expression=request_call)

        return cst.SimpleStatementLine(
            body=[
                cst.Assign(
                    targets=[cst.AssignTarget(target=cst.Name("_response"))],
                    value=request_call,
                )
            ]
        )

    def _get_success_response(
        self,
        operation: OperationMethod,
    ) -> tuple[str, OperationResponse] | None:
        """Get the first success response (2xx) from operation responses.

        Prioritizes: 200, 201, 202, then any other 2xx, then 'default'.
        """

        # Priority order for success codes
        priority_codes = ["200", "201", "202", "203", "204", "205", "206"]

        for code in priority_codes:
            if code in operation.responses:
                return (code, operation.responses[code])

        # Check for any other 2xx
        for code, response in operation.responses.items():
            if code.startswith("2"):
                return (code, response)

        # Fall back to 'default' if present
        if "default" in operation.responses:
            return ("default", operation.responses["default"])

        return None

    def _build_response_handling(
        self,
        operation: OperationMethod,
    ) -> list[cst.BaseStatement]:
        """Build response handling code."""
        statements: list[cst.BaseStatement] = []

        # For direct execution (no hooks), httpx response has .headers as Headers object
        # For hooked execution, we use ResponseData which has .headers as dict
        # Normalize headers to dict for consistent access
        if self._hooks_config is None:
            # Direct execution: convert headers to dict
            statements.append(
                cst.parse_statement("_headers_dict = dict(_response.headers)")
            )
            headers_expr: cst.BaseExpression = cst.Name("_headers_dict")
        else:
            # Hooked execution: headers already a dict in ResponseData
            headers_expr = cst.Attribute(
                value=cst.Name("_response"),
                attr=cst.Name("headers"),
            )

        # Check for error response
        statements.append(
            cst.If(
                test=cst.Comparison(
                    left=cst.Attribute(
                        value=cst.Name("_response"),
                        attr=cst.Name("status_code"),
                    ),
                    comparisons=[
                        cst.ComparisonTarget(
                            operator=cst.GreaterThanEqual(),
                            comparator=cst.Integer("400"),
                        )
                    ],
                ),
                body=cst.IndentedBlock(
                    body=[
                        cst.SimpleStatementLine(
                            body=[
                                cst.Raise(
                                    exc=cst.Call(
                                        func=cst.Name("APIError"),
                                        args=[
                                            cst.Arg(
                                                keyword=cst.Name("status_code"),
                                                value=cst.Attribute(
                                                    value=cst.Name("_response"),
                                                    attr=cst.Name("status_code"),
                                                ),
                                            ),
                                            cst.Arg(
                                                keyword=cst.Name("body"),
                                                value=cst.Attribute(
                                                    value=cst.Name("_response"),
                                                    attr=cst.Name("content"),
                                                ),
                                            ),
                                            cst.Arg(
                                                keyword=cst.Name("headers"),
                                                value=headers_expr,
                                            ),
                                        ],
                                    )
                                )
                            ]
                        )
                    ]
                ),
                orelse=None,
            )
        )

        # Get success response with priority handling
        success_result = self._get_success_response(operation)

        # Build decode and return statements
        decode_return_stmts: list[cst.BaseStatement] = []

        if success_result:
            status_code, success_response = success_result

            # Handle 204 No Content specially
            if status_code == "204":
                decode_return_stmts.append(self._build_return_none(headers_expr))
            elif success_response.type_annotation:
                # Decode response
                value_expr = self._build_decode_expression(
                    success_response.type_annotation,
                    success_response.content_type,
                )
                decode_return_stmts.append(
                    cst.SimpleStatementLine(
                        body=[
                            cst.Assign(
                                targets=[cst.AssignTarget(target=cst.Name("_result"))],
                                value=value_expr,
                            )
                        ]
                    )
                )

                # Apply on_result hook if configured
                if self._hooks_config is not None and self._hooks_config.on_result:
                    assert isinstance(self._hooks_config.on_result, str)
                    await_prefix = "await " if self.async_mode else ""
                    on_result_func = self._hooks_config.get_hook_function(
                        self._hooks_config.on_result
                    )
                    code = f"_result = {await_prefix}{on_result_func}(self._inner, _result, _info)"
                    decode_return_stmts.append(cst.parse_statement(code))

                decode_return_stmts.append(self._build_return_result(headers_expr))
            else:
                # No type annotation - return None
                decode_return_stmts.append(self._build_return_none(headers_expr))
        else:
            # No response defined - return None
            decode_return_stmts.append(self._build_return_none(headers_expr))

        # Wrap in try/except if on_error hook is configured
        if self._hooks_config is not None and self._hooks_config.on_error:
            assert isinstance(self._hooks_config.on_error, str)
            await_prefix = "await " if self.async_mode else ""
            on_error_func = self._hooks_config.get_hook_function(
                self._hooks_config.on_error
            )

            # Build except block
            except_code = f"""
except Exception as _e:
    {await_prefix}{on_error_func}(self._inner, _e, _info)
    raise
"""
            except_module = cst.parse_module(except_code.strip())
            except_handler = except_module.body[0]

            # Wrap decode/return in try block
            statements.append(
                cst.Try(
                    body=cst.IndentedBlock(body=decode_return_stmts),
                    handlers=[except_handler.handlers[0]],  # type: ignore[attr-defined]
                    orelse=None,
                    finalbody=None,
                )
            )
        else:
            statements.extend(decode_return_stmts)

        return statements

    def _build_return_result(
        self,
        headers_expr: cst.BaseExpression,
    ) -> cst.SimpleStatementLine:
        """Build return statement with ResponseValue wrapper using _result."""
        return cst.SimpleStatementLine(
            body=[
                cst.Return(
                    value=cst.Call(
                        func=cst.Name("ResponseValue"),
                        args=[
                            cst.Arg(
                                keyword=cst.Name("value"),
                                value=cst.Name("_result"),
                            ),
                            cst.Arg(
                                keyword=cst.Name("status_code"),
                                value=cst.Attribute(
                                    value=cst.Name("_response"),
                                    attr=cst.Name("status_code"),
                                ),
                            ),
                            cst.Arg(
                                keyword=cst.Name("headers"),
                                value=headers_expr,
                            ),
                        ],
                    )
                )
            ]
        )

    def _build_decode_expression(
        self,
        type_annotation: str,
        content_type: ContentType | None,
    ) -> cst.BaseExpression:
        """Build the decode expression based on content type."""
        # bytes expression: _response.content
        bytes_expr = cst.Attribute(
            value=cst.Name("_response"),
            attr=cst.Name("content"),
        )

        if content_type == ContentType.TEXT:
            # Decode bytes to text: _response.content.decode("utf-8")
            return cst.Call(
                func=cst.Attribute(
                    value=bytes_expr,
                    attr=cst.Name("decode"),
                ),
                args=[cst.Arg(value=cst.SimpleString('"utf-8"'))],
            )
        elif content_type == ContentType.BINARY:
            # Return bytes directly: _response.content
            return bytes_expr
        else:
            # JSON (default) - use dataclass backend's decode
            if self._dataclass_backend is not None:
                return self._dataclass_backend.generate_decode_json_expression(
                    type_annotation, bytes_expr
                )
            else:
                # Fallback to json.loads
                return cst.Call(
                    func=cst.Attribute(
                        value=cst.Name("json"),
                        attr=cst.Name("loads"),
                    ),
                    args=[cst.Arg(value=bytes_expr)],
                )

    def _build_return_value(
        self,
        value_expr: cst.BaseExpression,
    ) -> cst.SimpleStatementLine:
        """Build return statement with ResponseValue wrapper."""
        return cst.SimpleStatementLine(
            body=[
                cst.Return(
                    value=cst.Call(
                        func=cst.Name("ResponseValue"),
                        args=[
                            cst.Arg(
                                keyword=cst.Name("value"),
                                value=value_expr,
                            ),
                            cst.Arg(
                                keyword=cst.Name("status_code"),
                                value=cst.Attribute(
                                    value=cst.Name("_response"),
                                    attr=cst.Name("status_code"),
                                ),
                            ),
                            cst.Arg(
                                keyword=cst.Name("headers"),
                                value=cst.Attribute(
                                    value=cst.Name("_response"),
                                    attr=cst.Name("headers"),
                                ),
                            ),
                        ],
                    )
                )
            ]
        )

    def _build_return_none(
        self,
        headers_expr: cst.BaseExpression,
    ) -> cst.SimpleStatementLine:
        """Build return statement with ResponseValue[None]."""
        return cst.SimpleStatementLine(
            body=[
                cst.Return(
                    value=cst.Call(
                        func=cst.Name("ResponseValue"),
                        args=[
                            cst.Arg(
                                keyword=cst.Name("value"),
                                value=cst.Name("None"),
                            ),
                            cst.Arg(
                                keyword=cst.Name("status_code"),
                                value=cst.Attribute(
                                    value=cst.Name("_response"),
                                    attr=cst.Name("status_code"),
                                ),
                            ),
                            cst.Arg(
                                keyword=cst.Name("headers"),
                                value=headers_expr,
                            ),
                        ],
                    )
                )
            ]
        )

    def _get_return_type(self, operation: OperationMethod) -> str:
        """Get the return type annotation for an operation."""
        success_result = self._get_success_response(operation)

        if success_result:
            status_code, success_response = success_result

            # 204 No Content always returns None
            if status_code == "204":
                return "ResponseValue[None]"

            if success_response.type_annotation:
                return f"ResponseValue[{success_response.type_annotation}]"

        return "ResponseValue[None]"

    def get_imports(self) -> list[cst.SimpleStatementLine]:
        """Return required imports for httpx."""
        return [
            cst.SimpleStatementLine(
                body=[cst.Import(names=[cst.ImportAlias(name=cst.Name("httpx"))])]
            ),
        ]

    def supports_async(self) -> bool:
        """httpx supports async."""
        return True

    def supports_sync(self) -> bool:
        """httpx supports sync."""
        return True

    def get_sub_client_init_args(self) -> list[tuple[str, str]]:
        """Return args for sub-client initialization."""
        return [("client", "self._client"), ("base_url", "self._base_url")]



"""HTTP backend implementations."""

from senzo.backends.http.aiohttp import AiohttpBackend
from senzo.backends.http.httpx import HttpxBackend
from senzo.backends.http.requests import RequestsBackend

__all__ = [
    "AiohttpBackend",
    "HttpxBackend",
    "RequestsBackend",
]



from __future__ import annotations
from enum import Enum
from pydantic import BaseModel, ConfigDict, Field
import datetime
import uuid
from typing import final, Literal


@final
class FormatTypes(BaseModel):
    """Tests various format types"""

    created_date: datetime.date | None = Field(default=None)
    created_datetime: datetime.datetime | None = Field(default=None)
    item_uuid: uuid.UUID | None = Field(default=None)


@final
class HttpRequest(BaseModel):
    """Tests snake_case conversion of HTTPRequest -> http_request"""

    model_config = ConfigDict(populate_by_name=True)
    created_at: datetime.datetime | None = Field(default=None, alias="createdAt")
    http_header: str | None = Field(default=None)


@final
class ItemWithInlineEnum(BaseModel):
    """Tests inline enum without explicit type"""

    kind: Literal["a", "b", "c"] | None = Field(default=None)


class NullableStatus(str, Enum):
    ON = "on"
    OFF = "off"


class Status(str, Enum):
    ACTIVE = "active"
    IN_PROGRESS = "in-progress"
    _100_ = "100%"
    COMPLETED = "COMPLETED"



from __future__ import annotations
from typing import final
import aiohttp
from pydantic import TypeAdapter
from typing import Any
from ._base import APIError, ResponseValue, encode_path, ResponseData
from .types import (
    BearerResponse,
    HeadersResponse,
    HttpMethodResponse,
    IpResponse,
    PostResponse,
    UserAgentResponse,
)


@final
class Client:
    def __init__(self, base_url: str) -> None:
        self._base_url = base_url
        self._session: aiohttp.ClientSession | None = None

    async def close(self) -> None:
        if self._session is not None:
            await self._session.close()
            self._session = None

    async def __aenter__(self) -> "Client":
        if self._session is None:
            self._session = aiohttp.ClientSession(base_url=self._base_url)
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: object,
    ) -> None:
        await self.close()

    def _ensure_session(self) -> aiohttp.ClientSession:
        if self._session is None:
            raise RuntimeError("Client must be used as async context manager")
        return self._session

    async def get_request(self) -> ResponseValue[HttpMethodResponse]:
        """Returns the GET request's data"""
        _path = "/get"
        _url = self._base_url.rstrip("/") + _path
        _session = self._ensure_session()
        async with _session.request(
            method="GET",
            url=_url,
            headers={},
            params={},
            data=None,
        ) as _resp:
            _content = await _resp.read()
            _response = ResponseData(
                status_code=_resp.status,
                headers=dict(_resp.headers),
                content=_content,
            )
        if _response.status_code >= 400:
            raise APIError(
                status_code=_response.status_code,
                body=_response.content,
                headers=_response.headers,
            )
        _result = HttpMethodResponse.model_validate_json(_response.content)
        return ResponseValue(
            value=_result, status_code=_response.status_code, headers=_response.headers
        )

    async def post_request(
        self, body: dict[str, Any] | None = None
    ) -> ResponseValue[PostResponse]:
        """Returns the POSTed data"""
        _path = "/post"
        _url = self._base_url.rstrip("/") + _path
        _session = self._ensure_session()
        async with _session.request(
            method="POST",
            url=_url,
            headers={"Content-Type": "application/json"},
            params={},
            data=TypeAdapter(dict[str, Any]).dump_json(body)
            if body is not None
            else None,
        ) as _resp:
            _content = await _resp.read()
            _response = ResponseData(
                status_code=_resp.status,
                headers=dict(_resp.headers),
                content=_content,
            )
        if _response.status_code >= 400:
            raise APIError(
                status_code=_response.status_code,
                body=_response.content,
                headers=_response.headers,
            )
        _result = PostResponse.model_validate_json(_response.content)
        return ResponseValue(
            value=_result, status_code=_response.status_code, headers=_response.headers
        )

    async def get_origin_ip(self) -> ResponseValue[IpResponse]:
        """Returns Origin IP"""
        _path = "/ip"
        _url = self._base_url.rstrip("/") + _path
        _session = self._ensure_session()
        async with _session.request(
            method="GET",
            url=_url,
            headers={},
            params={},
            data=None,
        ) as _resp:
            _content = await _resp.read()
            _response = ResponseData(
                status_code=_resp.status,
                headers=dict(_resp.headers),
                content=_content,
            )
        if _response.status_code >= 400:
            raise APIError(
                status_code=_response.status_code,
                body=_response.content,
                headers=_response.headers,
            )
        _result = IpResponse.model_validate_json(_response.content)
        return ResponseValue(
            value=_result, status_code=_response.status_code, headers=_response.headers
        )

    async def get_user_agent(self) -> ResponseValue[UserAgentResponse]:
        """Returns the user agent"""
        _path = "/user-agent"
        _url = self._base_url.rstrip("/") + _path
        _session = self._ensure_session()
        async with _session.request(
            method="GET",
            url=_url,
            headers={},
            params={},
            data=None,
        ) as _resp:
            _content = await _resp.read()
            _response = ResponseData(
                status_code=_resp.status,
                headers=dict(_resp.headers),
                content=_content,
            )
        if _response.status_code >= 400:
            raise APIError(
                status_code=_response.status_code,
                body=_response.content,
                headers=_response.headers,
            )
        _result = UserAgentResponse.model_validate_json(_response.content)
        return ResponseValue(
            value=_result, status_code=_response.status_code, headers=_response.headers
        )

    async def get_headers(self) -> ResponseValue[HeadersResponse]:
        """Returns the request headers"""
        _path = "/headers"
        _url = self._base_url.rstrip("/") + _path
        _session = self._ensure_session()
        async with _session.request(
            method="GET",
            url=_url,
            headers={},
            params={},
            data=None,
        ) as _resp:
            _content = await _resp.read()
            _response = ResponseData(
                status_code=_resp.status,
                headers=dict(_resp.headers),
                content=_content,
            )
        if _response.status_code >= 400:
            raise APIError(
                status_code=_response.status_code,
                body=_response.content,
                headers=_response.headers,
            )
        _result = HeadersResponse.model_validate_json(_response.content)
        return ResponseValue(
            value=_result, status_code=_response.status_code, headers=_response.headers
        )

    async def get_status_code(self, status_code: int) -> ResponseValue[None]:
        """Returns the specified HTTP status code"""
        _path = f"/status/{encode_path(status_code)}"
        _url = self._base_url.rstrip("/") + _path
        _session = self._ensure_session()
        async with _session.request(
            method="GET",
            url=_url,
            headers={},
            params={},
            data=None,
        ) as _resp:
            _content = await _resp.read()
            _response = ResponseData(
                status_code=_resp.status,
                headers=dict(_resp.headers),
                content=_content,
            )
        if _response.status_code >= 400:
            raise APIError(
                status_code=_response.status_code,
                body=_response.content,
                headers=_response.headers,
            )
        return ResponseValue(
            value=None, status_code=_response.status_code, headers=_response.headers
        )

    async def get_delay(self, n: int) -> ResponseValue[HttpMethodResponse]:
        """Delays responding for n seconds"""
        _path = f"/delay/{encode_path(n)}"
        _url = self._base_url.rstrip("/") + _path
        _session = self._ensure_session()
        async with _session.request(
            method="GET",
            url=_url,
            headers={},
            params={},
            data=None,
        ) as _resp:
            _content = await _resp.read()
            _response = ResponseData(
                status_code=_resp.status,
                headers=dict(_resp.headers),
                content=_content,
            )
        if _response.status_code >= 400:
            raise APIError(
                status_code=_response.status_code,
                body=_response.content,
                headers=_response.headers,
            )
        _result = HttpMethodResponse.model_validate_json(_response.content)
        return ResponseValue(
            value=_result, status_code=_response.status_code, headers=_response.headers
        )

    async def test_bearer_auth(self) -> ResponseValue[BearerResponse]:
        """Tests Bearer authentication"""
        _path = "/bearer"
        _url = self._base_url.rstrip("/") + _path
        _session = self._ensure_session()
        async with _session.request(
            method="GET",
            url=_url,
            headers={},
            params={},
            data=None,
        ) as _resp:
            _content = await _resp.read()
            _response = ResponseData(
                status_code=_resp.status,
                headers=dict(_resp.headers),
                content=_content,
            )
        if _response.status_code >= 400:
            raise APIError(
                status_code=_response.status_code,
                body=_response.content,
                headers=_response.headers,
            )
        _result = BearerResponse.model_validate_json(_response.content)
        return ResponseValue(
            value=_result, status_code=_response.status_code, headers=_response.headers
        )



"""Configuration types for code generation."""

from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from typing import TYPE_CHECKING, Any, Callable, TypeAlias, TypedDict, TypeVar

_T = TypeVar("_T")

if TYPE_CHECKING:
    from senzo.backends.base import DataclassBackend, HttpBackend


# Type aliases for hook function signatures
PreHookFunc: TypeAlias = Callable[
    [_T, Any, Any], Any
]  # (inner, request, info) -> request
PostHookFunc: TypeAlias = Callable[
    [_T, Any, Any], Any
]  # (inner, response, info) -> response
OnErrorFunc: TypeAlias = Callable[
    [_T, Exception, Any], None
]  # (inner, error, info) -> None
OnResultFunc: TypeAlias = Callable[
    [_T, Any, Any], Any
]  # (inner, result, info) -> result


def _to_import_path(obj: str | type | Callable[..., Any]) -> str:
    """Convert a type or callable to its import path string."""
    if isinstance(obj, str):
        return obj
    return f"{obj.__module__}.{obj.__name__}"


@dataclass(slots=True)
class HooksConfig:
    """Configuration for generation-time hooks.

    Accepts either string paths (e.g., "myapi.auth.sign_request") or actual
    types/callables which are converted to paths automatically.

    Attributes:
        inner_type: The type class for client state, as a string path or type.
                    This type is passed to Client.__init__ and forwarded to all hooks.
        pre_hook: Optional pre-request hook function path or callable.
                  Signature: (inner: InnerType, request: RequestData, info: OperationInfo) -> RequestData
        post_hook: Optional post-response hook function path or callable.
                   Signature: (inner: InnerType, response: ResponseData, info: OperationInfo) -> ResponseData
        on_error: Optional error hook function path or callable.
                  Signature: (inner: InnerType, error: Exception, info: OperationInfo) -> None
        on_result: Optional result transformation hook function path or callable.
                   Signature: (inner: InnerType, result: T, info: OperationInfo) -> T
    """

    inner_type: str | type[Any]
    pre_hook: str | PreHookFunc[Any] | None = None
    post_hook: str | PostHookFunc[Any] | None = None
    on_error: str | OnErrorFunc[Any] | None = None
    on_result: str | OnResultFunc[Any] | None = None

    def __post_init__(self) -> None:
        # Convert types/callables to string paths
        self.inner_type = _to_import_path(self.inner_type)
        for attr in ("pre_hook", "post_hook", "on_error", "on_result"):
            val = getattr(self, attr)
            if val is not None and not isinstance(val, str):
                setattr(self, attr, _to_import_path(val))

    def get_inner_module(self) -> str:
        """Get the module path for the inner type."""
        assert isinstance(self.inner_type, str)
        return ".".join(self.inner_type.rsplit(".", 1)[:-1])

    def get_inner_class(self) -> str:
        """Get the class name for the inner type."""
        assert isinstance(self.inner_type, str)
        return self.inner_type.rsplit(".", 1)[-1]

    def get_hook_module(self, hook_path: str) -> str:
        """Get the module path for a hook function."""
        return ".".join(hook_path.rsplit(".", 1)[:-1])

    def get_hook_function(self, hook_path: str) -> str:
        """Get the function name for a hook."""
        return hook_path.rsplit(".", 1)[-1]


class PaginationConfig(TypedDict, total=False):
    """Configuration for a single paginated operation."""

    items: str  # Response field containing items array (default: "items")
    next_token: str  # Response field containing next token (default: "next_page")
    token_param: str  # Query param name for page token (default: "page_token")
    limit_param: str  # Query param name for limit (default: "limit")
    token_type: str  # Type of token: "cursor" (default) or "url"


class TagStyle(Enum):
    """How to handle operation tags in the generated client.

    - FLAT: All operations in a single Client class with original method names (default)
    - FLAT_PREFIXED: All operations in a single Client with tag-prefixed method names
      (e.g., users_get_user, orders_create_order)
    - GROUPED: Separate client classes per tag with main client aggregating sub-clients
      (e.g., client.users.get_user(), client.orders.create_order())
    """

    FLAT = "flat"
    FLAT_PREFIXED = "flat_prefixed"
    GROUPED = "grouped"

    # Backwards-compatible aliases (deprecated)
    MERGED = "flat"
    SEPARATE = "grouped"


class EnumStyle(Enum):
    """How to generate enum types.

    - LITERAL: Use Literal["a", "b", "c"] union types (default for inline enums)
    - ENUM: Generate enum.Enum classes (Python 3.4+)
    - STR_ENUM: Generate StrEnum classes (Python 3.11+, recommended)
    - STR: Fall back to plain str type (no type safety)
    """

    LITERAL = "literal"
    ENUM = "enum"
    STR_ENUM = "str_enum"
    STR = "str"


class GeneratorConfig(TypedDict, total=False):
    """Configuration options for code generation."""

    package_name: str
    dataclass_backend: DataclassBackend
    http_backend: HttpBackend
    tag_style: TagStyle
    enum_style: EnumStyle  # For inline enums in TypeSpace
    pagination: dict[str, PaginationConfig]  # operation_id -> pagination config
    hooks: HooksConfig



"""Code generation utilities."""

from senzo.codegen.utils import (
    make_annotation,
    make_assign,
    make_docstring,
    make_import_from,
    make_module,
    make_name,
    make_simple_string,
)

__all__ = [
    "make_annotation",
    "make_assign",
    "make_docstring",
    "make_import_from",
    "make_module",
    "make_name",
    "make_simple_string",
]



"""Kalshi authentication hooks for generation-time integration.

This module provides authentication state and hook functions that can be
baked into generated clients via HooksConfig.
"""

from __future__ import annotations

import base64
import time
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any
from urllib.parse import urlparse

if TYPE_CHECKING:
    pass


@dataclass
class KalshiAuthState:
    """State for Kalshi RSA-PSS authentication."""

    api_key_id: str
    private_key: Any  # RSAPrivateKey

    def sign(self, message: str) -> str:
        """Sign a message with RSA-PSS."""
        from cryptography.hazmat.primitives import hashes
        from cryptography.hazmat.primitives.asymmetric import padding

        signature = self.private_key.sign(
            message.encode("utf-8"),
            padding.PSS(
                mgf=padding.MGF1(hashes.SHA256()),
                salt_length=padding.PSS.MAX_LENGTH,
            ),
            hashes.SHA256(),
        )
        return base64.b64encode(signature).decode("utf-8")


async def sign_request(inner: KalshiAuthState, request: Any, info: Any) -> Any:
    """Async pre-request hook that adds Kalshi auth headers.

    Args:
        inner: The KalshiAuthState instance with credentials
        request: RequestData from the generated client's _base module
        info: OperationInfo from the generated client's _base module

    Returns:
        New RequestData with authentication headers added
    """
    timestamp_ms = str(int(time.time() * 1000))
    parsed = urlparse(str(request.url))
    message = f"{timestamp_ms}{request.method}{parsed.path}"
    signature = inner.sign(message)

    new_headers = dict(request.headers)
    new_headers["KALSHI-ACCESS-KEY"] = inner.api_key_id
    new_headers["KALSHI-ACCESS-TIMESTAMP"] = timestamp_ms
    new_headers["KALSHI-ACCESS-SIGNATURE"] = signature

    return request.__class__(
        method=request.method,
        url=request.url,
        headers=new_headers,
        params=request.params,
        content=request.content,
    )


def sign_request_sync(inner: KalshiAuthState, request: Any, info: Any) -> Any:
    """Sync pre-request hook that adds Kalshi auth headers.

    Args:
        inner: The KalshiAuthState instance with credentials
        request: RequestData from the generated client's _base module
        info: OperationInfo from the generated client's _base module

    Returns:
        New RequestData with authentication headers added
    """
    timestamp_ms = str(int(time.time() * 1000))
    parsed = urlparse(str(request.url))
    message = f"{timestamp_ms}{request.method}{parsed.path}"
    signature = inner.sign(message)

    new_headers = dict(request.headers)
    new_headers["KALSHI-ACCESS-KEY"] = inner.api_key_id
    new_headers["KALSHI-ACCESS-TIMESTAMP"] = timestamp_ms
    new_headers["KALSHI-ACCESS-SIGNATURE"] = signature

    return request.__class__(
        method=request.method,
        url=request.url,
        headers=new_headers,
        params=request.params,
        content=request.content,
    )



"""Runtime support for the generated client."""

from __future__ import annotations

from collections.abc import Awaitable, Callable
from dataclasses import dataclass
from typing import Any, Generic, TypeVar, final

_T = TypeVar("_T")


@dataclass(frozen=True, slots=True)
class ResponseValue(Generic[_T]):
    """Wrapper for successful API responses."""

    value: _T
    status_code: int
    headers: dict[str, str]


@dataclass(frozen=True, slots=True)
class APIError(Exception):
    """Exception for API errors."""

    status_code: int
    body: bytes
    headers: dict[str, str] | None = None

    def json(self) -> Any:
        """Parse error body as JSON."""
        import json

        return json.loads(self.body)


def encode_path(value: str | int) -> str:
    """URL-encode a path parameter value."""
    from urllib.parse import quote

    return quote(str(value), safe="")


def extract_param_from_url(url: str | None, param: str) -> str | None:
    """Extract a query parameter value from a URL."""
    if not url:
        return None
    from urllib.parse import urlparse, parse_qs

    qs = parse_qs(urlparse(url).query)
    values = qs.get(param)
    return values[0] if values else None


@dataclass(frozen=True, slots=True)
class RequestData:
    """Request data passed to hooks."""

    method: str
    url: str
    headers: dict[str, str]
    params: dict[str, Any]
    content: bytes | None


@dataclass(frozen=True, slots=True)
class ResponseData:
    """Response data passed to hooks."""

    status_code: int
    headers: dict[str, str]
    content: bytes


@dataclass(frozen=True, slots=True)
class OperationInfo:
    """Operation metadata passed to hooks."""

    operation_id: str
    method: str
    path_template: str
    tags: list[str]
    security: list[str]


_PageT = TypeVar("_PageT")


@final
class Paginator(Generic[_T]):
    """Generic async paginator with convenience methods."""

    def __init__(
        self,
        fetch_page: Callable[[str | None], Awaitable[ResponseValue[_PageT]]],
        get_items: Callable[[_PageT], list[_T]],
        get_next_token: Callable[[_PageT], str | None],
    ) -> None:
        self._fetch_page = fetch_page
        self._get_items = get_items
        self._get_next_token = get_next_token
        self._page_token: str | None = None
        self._buffer: list[_T] = []
        self._exhausted = False

    def __aiter__(self) -> "Paginator[_T]":
        return self

    async def __anext__(self) -> _T:
        while not self._buffer and not self._exhausted:
            response = await self._fetch_page(self._page_token)
            page = response.value
            self._buffer = list(self._get_items(page))
            self._page_token = self._get_next_token(page)
            if not self._page_token:
                self._exhausted = True

        if not self._buffer:
            raise StopAsyncIteration

        return self._buffer.pop(0)

    async def atake(self, n: int) -> list[_T]:
        """Take up to n items."""
        result: list[_T] = []
        async for item in self:
            result.append(item)
            if len(result) >= n:
                break
        return result
